Welcome to {{ aux.python_project.second_name }}'s documentation!
==========={{(aux.python_project.second_name|length)*"="}}=================
.. toctree::
   :maxdepth: 2
   :caption: Contents:

   modules

Indices and tables
==================

* :ref:`genindex`
* :ref:`modindex`
* :ref:`search`
