# automatically generated by auxilium
{%- if "push" in aux.ci.used_rules %}
.r-push:
    if: '$CI_PIPELINE_SOURCE == "push" && $CI_COMMIT_TAG == null'
{% endif -%}
{%- if "push_no_mr" in aux.ci.used_rules %}
.r-push_no_mr:
    if: '$CI_PIPELINE_SOURCE == "push" && $CI_OPEN_MERGE_REQUESTS == null && $CI_COMMIT_TAG == null'
{% endif -%}
{%- for key, (_, rule) in aux.ci.used_tag_rules.items() %}
.r-{{ key }}:
    if: '{{ rule }}'
{% endfor -%}
{%- if "web" in aux.ci.used_rules %}
.r-web:
    if: $CI_PIPELINE_SOURCE == "web"
{% endif -%}
{%- if "mr" in aux.ci.used_rules %}
.r-mr:
    if: $CI_PIPELINE_SOURCE == "merge_request_event"
{% endif -%}
{%- if "pipeline" in aux.ci.used_rules %}
.r-pipeline:
    if: $CI_PIPELINE_SOURCE == "pipeline"
{% endif -%}
{%- if "schedule" in aux.ci.used_rules %}
.r-schedule:
    if: $CI_PIPELINE_SOURCE == "schedule"
{% endif -%}
{%- if "docs" not in aux.meta.disable %}
.r-release-push:
    if: '$CI_COMMIT_BRANCH == "{{aux.gitlab.release_branch}}" && $CI_COMMIT_TAG == null'
{% endif -%}
